from importlib.metadata import version
from os import path
from tempfile import TemporaryDirectory
from typing import Iterable, Iterator, Optional
from urllib.request import Request, urlopen, urlretrieve

from camelot import read_pdf  # type: ignore
from bs4 import BeautifulSoup

from ironbot.models import Athlete, Event, Title
from ironbot.parsers import EventParser


URL = "https://www.ironman.com/pro-athletes"
CALENDAR = "PRO Schedule".lower()
HEADERS = {"User-Agent": f"ironbot/{version('ironbot')}", "Accept": "*/*"}


def pdf_table_rows(url: str) -> Iterable[Iterator[str]]:
    with TemporaryDirectory() as tmp_dir:
        tmp = path.join(tmp_dir, "tmp.pdf")
        urlretrieve(url, tmp)

        tables = read_pdf(tmp, pages="all")
        if not tables:
            tables = read_pdf(tmp, pages="all", flavor="stream")

        for table in tables:
            yield from table.df.itertuples(index=False, name=None)


def load(title: Title) -> BeautifulSoup:
    req = Request(URL, headers=HEADERS)
    with urlopen(req) as resp:
        dom = BeautifulSoup(resp, features="html.parser")
        for h3 in dom.find_all("h3"):
            if h3.text != title:
                continue

            return h3.parent

    raise RuntimeError(f"HTML block not found for {title}")


def events(data: BeautifulSoup) -> Iterable[Event]:
    for a in data.find_all("a"):
        if a.text.strip().lower() != CALENDAR:
            continue

        for row in pdf_table_rows(a["href"]):
            yield from EventParser(row)

        return

    raise RuntimeError("Calendar URL not found")


def event_names(data: BeautifulSoup) -> Iterable[str]:
    yield from (a.text for a in data.find_all("a"))


def start_list(data: BeautifulSoup, event_number: int) -> Iterable[Athlete]:
    try:
        link = data.find_all("a")[event_number - 1]
    except IndexError:
        raise RuntimeError("Start list URL not found")

    for row in pdf_table_rows(link["href"]):
        try:
            yield Athlete.from_row(row)
        except RuntimeError:
            pass



from datetime import datetime
from enum import Enum
from re import match
from typing import Iterator

from sqlalchemy import Column, Integer, String, Date
from sqlalchemy.orm import DeclarativeBase


CATEGORY = r"^[MWF](PRO)?$"


class Title(Enum):
    CALENDAR: str = "PRO Event Calendar"
    START_LIST: str = "Start Lists"

    def __eq__(self, other) -> bool:
        return self.value.lower() == other.strip().lower()


class Base(DeclarativeBase):
    pass


class Event(Base):
    __tablename__ = "events"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    when = Column(Date)
    prize = Column(String)
    slots = Column(String)
    registration = Column(String)
    deadline = Column(String)

    def __init__(self, *args, **kwargs):
        kwargs["when"] = datetime.strptime(kwargs["when"], "%m/%d/%Y").date()
        return super().__init__(*args, **kwargs)

    def __str__(self) -> str:
        fields = (
            self.when.strftime("%Y-%m-%d"),
            self.name,
            self.prize,
            self.slots,
            self.registration,
            self.deadline,
        )
        return "\t".join(str(field) for field in fields)

    def __repr__(self) -> str:
        return str(self)


class Athlete(Base):
    __tablename__ = "athletes"

    id = Column(Integer, primary_key=True)
    bib = Column(String)
    first_name = Column(String)
    last_name = Column(String)
    country = Column(String)
    category = Column(String)

    @classmethod
    def from_row(cls, row: Iterator[str]) -> "Athlete":
        fields = {column.name: "" for column in cls.__table__.columns}

        for field in (field.strip() for field in row):
            if field.isnumeric() and not fields["bib"]:
                fields["bib"] = field
            elif match(CATEGORY, field.upper()) and not fields["category"]:
                fields["category"] = field
            elif not fields["last_name"]:
                fields["last_name"] = field
            elif not fields["first_name"]:
                fields["first_name"] = field
            elif not fields["country"]:
                fields["country"] = field

        if not all(fields[key] for key in ("bib", "first_name", "last_name")):
            raise RuntimeError(f"Could not parse athlete from: {row}")

        return cls(**fields)

    @property
    def name(self) -> str:
        return f"{self.first_name} {self.last_name}"

    def __str__(self) -> str:
        fields = (self.bib, self.category, self.name, self.country)
        return "\t".join(str(field) for field in fields if field)

    def __repr__(self) -> str:
        return str(self)



from typing import Generator, Iterable, Optional

from ironbot.models import Event


class EventParser:
    COLUMNS = ("when", "name", "prize", "slots", "registration", "deadline")

    def __init__(self, row: Iterable[str]):
        self.original_row = row

    def parse(self, row: Iterable[str]) -> Optional[Event]:
        kwargs = dict(zip(self.COLUMNS, row))
        try:
            return Event(**kwargs)
        except (TypeError, KeyError, ValueError):
            return None

    def _events(self) -> Iterable[Optional[Event]]:
        when, *_ = self.original_row
        if "\n" not in when:
            yield self.parse(self.original_row)
            return

        for details in when.split("\n"):
            new_row = list(self.original_row)
            when, category = details.split(" ", 1)
            new_row[0] = when.strip()
            new_row[1] = f"{new_row[1]} {category.strip()}"
            yield self.parse(new_row)

    def __iter__(self) -> Generator[Event, None, None]:
        yield from (event for event in self._events() if event)
